@@ -12,6 +12,16 @@
     :license: MIT, see LICENSE for more details
 """
 
+from typing import (
+    Any,
+    Callable,
+    Iterable,
+    List,
+    Optional,
+    Set,
+    Union,
+)
+
 import functools
 import logging
 import socket as Socket
@@ -19,6 +29,8 @@ import ssl
 import weakref
 import uuid
 
+from asyncio import iscoroutinefunction, wait, Future
+
 import xml.etree.ElementTree as ET
 
 from slixmpp.xmlstream.asyncio import asyncio
@@ -30,6 +42,10 @@ from slixmpp.xmlstream.resolver import resolve, default_resolver
 RESPONSE_TIMEOUT = 30
 
 log = logging.getLogger(__name__)
+class ContinueQueue(Exception):
+    """
+    Exception raised in the send queue to "continue" from within an inner loop
+    """
 
 class NotConnectedError(Exception):
     """
@@ -81,6 +97,8 @@ class XMLStream(asyncio.BaseProtocol):
         self.force_starttls = None
         self.disable_starttls = None
 
+        self.waiting_queue = asyncio.Queue()
+
         # A dict of {name: handle}
         self.scheduled_events = {}
 
@@ -199,11 +217,6 @@ class XMLStream(asyncio.BaseProtocol):
         self.__event_handlers = {}
         self.__filters = {'in': [], 'out': [], 'out_sync': []}
 
-        self._id = 0
-
-        #: We use an ID prefix to ensure that all ID values are unique.
-        self._id_prefix = '%s-' % uuid.uuid4()
-
         # Current connection attempt (Future)
         self._current_connection_attempt = None
 
@@ -215,12 +228,18 @@ class XMLStream(asyncio.BaseProtocol):
         #: ``_xmpp-client._tcp`` service.
         self.dns_service = None
 
+        #: The reason why we are disconnecting from the server
+        self.disconnect_reason = None
+
         #: An asyncio Future being done when the stream is disconnected.
-        self.disconnected = asyncio.Future()
+        self.disconnected: Future = Future()
 
         self.add_event_handler('disconnected', self._remove_schedules)
         self.add_event_handler('session_start', self._start_keepalive)
 
+        self._run_out_filters: Optional[Future] = None
+        self.__slow_tasks: List[Future] = []
+
     @property
     def loop(self):
         if self._loop is None:
@@ -238,12 +257,13 @@ class XMLStream(asyncio.BaseProtocol):
         ID values. Using this method ensures that all new ID values
         are unique in this stream.
         """
-        self._id += 1
-        return self.get_id()
+        return uuid.uuid4().hex
 
-    def get_id(self):
-        """Return the current unique stream ID in hexadecimal form."""
-        return "%s%X" % (self._id_prefix, self._id)
+    def _set_disconnected_future(self):
+        """Set the self.disconnected future on disconnect"""
+        if not self.disconnected.done():
+            self.disconnected.set_result(True)
+        self.disconnected = asyncio.Future()
 
     def connect(self, host='', port=0, use_ssl=False,
                 force_starttls=True, disable_starttls=False):
@@ -256,7 +276,6 @@ class XMLStream(asyncio.BaseProtocol):
                         connection will be upgraded to SSL/TLS later, using
                         STARTTLS.  Only use this value for old servers that
                         have specific port for SSL/TLS
-        TODO fix the comment
         :param force_starttls: If True, the connection will be aborted if
                                the server does not initiate a STARTTLS
                                negotiation.  If None, the connection will be
@@ -268,7 +287,15 @@ class XMLStream(asyncio.BaseProtocol):
                                localhost
 
         """
+        if self._run_out_filters is None or self._run_out_filters.done():
+            self._run_out_filters = asyncio.ensure_future(
+                self.run_filters(),
+                loop=self.loop,
+            )
+
+        self.disconnect_reason = None
         self.cancel_connection_attempt()
+        self.connect_loop_wait = 0
         if host and port:
             self.address = (host, int(port))
         try:
@@ -293,6 +320,10 @@ class XMLStream(asyncio.BaseProtocol):
     async def _connect_routine(self):
         self.event_when_connected = "connected"
 
+        if self.connect_loop_wait > 0:
+            self.event('reconnect_delay', self.connect_loop_wait)
+            await asyncio.sleep(self.connect_loop_wait, loop=self.loop)
+
         record = await self.pick_dns_answer(self.default_domain)
         if record is not None:
             host, address, dns_port = record
@@ -309,7 +340,8 @@ class XMLStream(asyncio.BaseProtocol):
         else:
             ssl_context = None
 
-        await asyncio.sleep(self.connect_loop_wait, loop=self.loop)
+        if self._current_connection_attempt is None:
+            return
         try:
             await self.loop.create_connection(lambda: self,
                                                    self.address[0],
@@ -323,6 +355,8 @@ class XMLStream(asyncio.BaseProtocol):
         except OSError as e:
             log.debug('Connection failed: %s', e)
             self.event("connection_failed", e)
+            if self._current_connection_attempt is None:
+                return
             self.connect_loop_wait = self.connect_loop_wait * 2 + 1
             self._current_connection_attempt = asyncio.ensure_future(
                 self._connect_routine(),
@@ -364,6 +398,7 @@ class XMLStream(asyncio.BaseProtocol):
             "ssl_object",
             default=self.transport.get_extra_info("socket")
         )
+        self._current_connection_attempt = None
         self.init_parser()
         self.send_raw(self.stream_header)
         self.dns_answers = None
@@ -399,7 +434,9 @@ class XMLStream(asyncio.BaseProtocol):
                         # The stream's root element has closed,
                         # terminating the stream.
                         log.debug("End of stream received")
+                        self.disconnect_reason = "End of stream"
                         self.abort()
+                        return
                     elif self.xml_depth == 1:
                         # A stanza is an XML element that is a direct child of
                         # the root element, hence the check of depth == 1
@@ -420,6 +457,9 @@ class XMLStream(asyncio.BaseProtocol):
             self.send(error)
             self.disconnect()
 
+    def is_connecting(self):
+        return self._current_connection_attempt is not None
+
     def is_connected(self):
         return self.transport is not None
 
@@ -433,14 +473,17 @@ class XMLStream(asyncio.BaseProtocol):
         closure of the TCP connection
         """
         log.info("connection_lost: %s", (exception,))
-        self.event("disconnected")
-        if self.end_session_on_disconnect:
-            self.event('session_end')
         # All these objects are associated with one TCP connection.  Since
         # we are not connected anymore, destroy them
         self.parser = None
         self.transport = None
         self.socket = None
+        # Fire the events after cleanup
+        if self.end_session_on_disconnect:
+            self._reset_sendq()
+            self.event('session_end')
+        self._set_disconnected_future()
+        self.event("disconnected", self.disconnect_reason or exception and exception.strerror)
 
     def cancel_connection_attempt(self):
         """
@@ -453,44 +496,100 @@ class XMLStream(asyncio.BaseProtocol):
             self._current_connection_attempt.cancel()
             self._current_connection_attempt = None
 
-    def disconnect(self, wait=2.0):
+    def disconnect(self, wait: Union[float, int] = 2.0, reason: Optional[str] = None, ignore_send_queue: bool = False) -> Future:
         """Close the XML stream and wait for an acknowldgement from the server for
         at most `wait` seconds.  After the given number of seconds has
-        passed without a response from the serveur, or when the server
+        passed without a response from the server, or when the server
         successfully responds with a closure of its own stream, abort() is
-        called. If wait is 0.0, this is almost equivalent to calling abort()
-        directly.
+        called. If wait is 0.0, this will call abort() directly without closing
+        the stream.
 
-        Does nothing if we are not connected.
+        Does nothing but trigger the disconnected event if we are not connected.
 
         :param wait: Time to wait for a response from the server.
-
+        :param reason: An optional reason for the disconnect.
+        :param ignore_send_queue: Boolean to toggle if we want to ignore
+                                  the in-flight stanzas and disconnect immediately.
+        :return: A future that ends when all code involved in the disconnect has ended
         """
-        self.cancel_connection_attempt()
+        # Compat: docs/getting_started/sendlogout.rst has been promoting
+        # `disconnect(wait=True)` for ages. This doesn't mean anything to the
+        # schedule call below. It would fortunately be converted to `1` later
+        # down the call chain. Praise the implicit casts lord.
+        if wait == True:
+            wait = 2.0
+
         if self.transport:
+            self.disconnect_reason = reason
+            if self.waiting_queue.empty() or ignore_send_queue:
+                self.cancel_connection_attempt()
+                return asyncio.ensure_future(
+                    self._end_stream_wait(wait, reason=reason),
+                    loop=self.loop,
+                )
+            else:
+                return asyncio.ensure_future(
+                    self._consume_send_queue_before_disconnecting(reason, wait),
+                    loop=self.loop,
+                )
+        else:
+            self._set_disconnected_future()
+            self.event("disconnected", reason)
+            future = Future()
+            future.set_result(None)
+            return future
+
+    async def _consume_send_queue_before_disconnecting(self, reason: Optional[str], wait: float):
+        """Wait until the send queue is empty before disconnecting"""
+        try:
+            await asyncio.wait_for(
+                self.waiting_queue.join(),
+                wait,
+                loop=self.loop
+            )
+        except asyncio.TimeoutError:
+            wait = 0 # we already consumed the timeout
+        self.disconnect_reason = reason
+        await self._end_stream_wait(wait)
+
+    async def _end_stream_wait(self, wait: Union[int, float] = 2, reason: Optional[str] = None):
+        """
+        Run abort() if we do not received the disconnected event
+        after a waiting time.
+
+        :param wait: The waiting time (defaults to 2)
+        """
+        try:
             self.send_raw(self.stream_footer)
-            self.schedule('Disconnect wait', wait,
-                          self.abort, repeat=False)
+            await self.wait_until('disconnected', wait)
+        except asyncio.TimeoutError:
+            self.abort()
+        except NotConnectedError:
+            # We are not connected when sending the end of stream
+            # that means the disconnect has already been handled
+            pass
 
     def abort(self):
         """
         Forcibly close the connection
         """
-        self.cancel_connection_attempt()
         if self.transport:
+            self.cancel_connection_attempt()
             self.transport.close()
             self.transport.abort()
             self.event("killed")
-            self.disconnected.set_result(True)
-            self.disconnected = asyncio.Future()
 
-    def reconnect(self, wait=2.0):
+    def reconnect(self, wait=2.0, reason="Reconnecting"):
         """Calls disconnect(), and once we are disconnected (after the timeout, or
         when the server acknowledgement is received), call connect()
         """
         log.debug("reconnecting...")
-        self.disconnect(wait)
-        self.add_event_handler('disconnected', self.connect, disposable=True)
+        async def handler(event):
+            # We yield here to allow synchronous handlers to work first
+            await asyncio.sleep(0, loop=self.loop)
+            self.connect()
+        self.add_event_handler('disconnected', handler, disposable=True)
+        self.disconnect(wait, reason)
 
     def configure_socket(self):
         """Set timeout and other options for self.socket.
@@ -564,6 +663,10 @@ class XMLStream(asyncio.BaseProtocol):
             else:
                 self.event('ssl_invalid_chain', e)
             return False
+        except OSError as exc:
+            log.debug("Connection error:", exc_info=True)
+            self.disconnect()
+            return False
         der_cert = transp.get_extra_info("ssl_object").getpeercert(True)
         pem_cert = ssl.DER_cert_to_PEM_cert(der_cert)
         self.event('ssl_cert', pem_cert)
@@ -593,7 +696,6 @@ class XMLStream(asyncio.BaseProtocol):
     def _remove_schedules(self, event):
         """Remove some schedules that become pointless when disconnected"""
         self.cancel_schedule('Whitespace Keepalive')
-        self.cancel_schedule('Disconnect wait')
 
     def start_stream_handler(self, xml):
         """Perform any initialization actions, such as handshakes,
@@ -771,14 +873,14 @@ class XMLStream(asyncio.BaseProtocol):
         """
         log.debug("Event triggered: %s", name)
 
-        handlers = self.__event_handlers.get(name, [])
+        handlers = self.__event_handlers.get(name, [])[:]
         for handler in handlers:
             handler_callback, disposable = handler
             old_exception = getattr(data, 'exception', None)
 
             # If the callback is a coroutine, schedule it instead of
             # running it directly
-            if asyncio.iscoroutinefunction(handler_callback):
+            if iscoroutinefunction(handler_callback):
                 async def handler_callback_routine(cb):
                     try:
                         await cb(data)
@@ -865,7 +967,9 @@ class XMLStream(asyncio.BaseProtocol):
         Execute the callback and remove the handler for it.
         """
         self._safe_cb_run(name, cb)
-        del self.scheduled_events[name]
+        # workaround for specific events which unschedule themselves
+        if name in self.scheduled_events:
+            del self.scheduled_events[name]
 
     def incoming_filter(self, xml):
         """Filter incoming XML objects before they are processed.
@@ -877,11 +981,107 @@ class XMLStream(asyncio.BaseProtocol):
         """
         return xml
 
+    def _reset_sendq(self):
+        """Clear sending tasks on session end"""
+        # Cancel all pending slow send tasks
+        log.debug('Cancelling %d slow send tasks', len(self.__slow_tasks))
+        for slow_task in self.__slow_tasks:
+            slow_task.cancel()
+        self.__slow_tasks.clear()
+        # Purge pending stanzas
+        while not self.waiting_queue.empty():
+            discarded = self.waiting_queue.get_nowait()
+            log.debug('Discarded stanza: %s', discarded)
+
+    async def _continue_slow_send(
+            self,
+            task: asyncio.Task,
+            already_used: Set[Callable[[ElementBase], Optional[StanzaBase]]]
+    ) -> None:
+        """
+        Used when an item in the send queue has taken too long to process.
+
+        This is away from the send queue and can take as much time as needed.
+        :param asyncio.Task task: the Task wrapping the coroutine
+        :param set already_used: Filters already used on this outgoing stanza
+        """
+        data = await task
+        self.__slow_tasks.remove(task)
+        for filter in self.__filters['out']:
+            if filter in already_used:
+                continue
+            if iscoroutinefunction(filter):
+                data = await task
+            else:
+                data = filter(data)
+            if data is None:
+                return
+
+        if isinstance(data, ElementBase):
+            for filter in self.__filters['out_sync']:
+                data = filter(data)
+                if data is None:
+                    return
+            str_data = tostring(data.xml, xmlns=self.default_ns,
+                                stream=self, top_level=True)
+            self.send_raw(str_data)
+        else:
+            self.send_raw(data)
+
+    async def run_filters(self):
+        """
+        Background loop that processes stanzas to send.
+        """
+        while True:
+            (data, use_filters) = await self.waiting_queue.get()
+            try:
+                if isinstance(data, ElementBase):
+                    if use_filters:
+                        already_run_filters = set()
+                        for filter in self.__filters['out']:
+                            already_run_filters.add(filter)
+                            if iscoroutinefunction(filter):
+                                task = asyncio.create_task(filter(data))
+                                completed, pending = await wait(
+                                    {task},
+                                    timeout=1,
+                                )
+                                if pending:
+                                    self.slow_tasks.append(task)
+                                    asyncio.ensure_future(
+                                        self._continue_slow_send(
+                                            task,
+                                            already_run_filters
+                                        ),
+                                        loop=self.loop,
+                                    )
+                                    raise Exception("Slow coro, rescheduling")
+                                data = task.result()
+                            else:
+                                data = filter(data)
+                            if data is None:
+                                raise ContinueQueue('Empty stanza')
+
+                if isinstance(data, ElementBase):
+                    if use_filters:
+                        for filter in self.__filters['out_sync']:
+                            data = filter(data)
+                            if data is None:
+                                raise ContinueQueue('Empty stanza')
+                    str_data = tostring(data.xml, xmlns=self.default_ns,
+                                        stream=self, top_level=True)
+                    self.send_raw(str_data)
+                else:
+                    self.send_raw(data)
+            except ContinueQueue as exc:
+                log.debug('Stanza in send queue not sent: %s', exc)
+            except Exception:
+                log.error('Exception raised in send queue:', exc_info=True)
+            self.waiting_queue.task_done()
+
     def send(self, data, use_filters=True):
         """A wrapper for :meth:`send_raw()` for sending stanza objects.
 
-        May optionally block until an expected response is received.
-
         :param data: The :class:`~slixmpp.xmlstream.stanzabase.ElementBase`
                      stanza to send on the stream.
         :param bool use_filters: Indicates if outgoing filters should be
@@ -889,24 +1089,7 @@ class XMLStream(asyncio.BaseProtocol):
                                  filters is useful when resending stanzas.
                                  Defaults to ``True``.
         """
-        if isinstance(data, ElementBase):
-            if use_filters:
-                for filter in self.__filters['out']:
-                    data = filter(data)
-                    if data is None:
-                        return
-
-        if isinstance(data, ElementBase):
-            if use_filters:
-                for filter in self.__filters['out_sync']:
-                    data = filter(data)
-                    if data is None:
-                        return
-            str_data = tostring(data.xml, xmlns=self.default_ns,
-                                stream=self, top_level=True)
-            self.send_raw(str_data)
-        else:
-            self.send_raw(data)
+        self.waiting_queue.put_nowait((data, use_filters))
 
     def send_xml(self, data):
         """Send an XML object on the stream
@@ -1004,3 +1187,24 @@ class XMLStream(asyncio.BaseProtocol):
         :param exception: An unhandled exception object.
         """
         pass
+
+    async def wait_until(self, event: str, timeout=30) -> Any:
+        """Utility method to wake on the next firing of an event.
+        (Registers a disposable handler on it)
+
+        :param str event: Event to wait on.
+        :param int timeout: Timeout
+        """
+        fut = asyncio.Future()
+        def result_handler(event_data):
+            if not fut.done():
+                fut.set_result(event_data)
+            else:
+                log.debug("Future registered on event '%s' was alredy done", event)
+
+        self.add_event_handler(
+            event,
+            result_handler,
+            disposable=True,
+        )
+        return await asyncio.wait_for(fut, timeout, loop=self.loop)
